#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

/* MSDOSism */
#ifndef	O_BINARY
#define	O_BINARY	0
#endif

static char devicename[FILENAME_MAX] = "";
static char bootrom[FILENAME_MAX] = "BOOT.DSK";
static char disk_a[FILENAME_MAX] = "A.DSK";
static char disk_b[FILENAME_MAX] = "B.DSK";
static char disk_c[FILENAME_MAX] = "C.DSK";
static char disk_d[FILENAME_MAX] = "D.DSK";
static char disk_e[FILENAME_MAX] = "E.DSK";
static char disk_f[FILENAME_MAX] = "F.DSK";
static char disk_g[FILENAME_MAX] = "G.DSK";
static char disk_h[FILENAME_MAX] = "H.DSK";
static char nodename[16] = "NONAME";


static uint32_t lba0;	/* first sector of the CP/M image area */
static uint8_t buff[512];
static uint8_t mbr[512];
static uint8_t boot[256];

void write_long(uint8_t *buff, int offs, int val)
{
	buff[offs+0] = (uint8_t)(val);
	buff[offs+1] = (uint8_t)(val >> 8);
	buff[offs+2] = (uint8_t)(val >> 16);
	buff[offs+3] = (uint8_t)(val >> 24);
}

int get_base_lba(int fd)
{
	int bytes_read;
	int i;

	lseek(fd, 0, SEEK_SET);
	bytes_read = read(fd, buff, sizeof(buff));
	if (bytes_read != sizeof(buff)) {
		perror("read()");
		return -1;
	}
	for (i = 0; i < 48; i++) {
		if (i % 16 == 0)
			printf("%04x:", i);
		printf(" %02x", buff[i]);
		if (i % 16 == 15)
			printf("\n");
	}
	lba0 = ((uint32_t)buff[32]) +
		((uint32_t)buff[33] << 8) +
		((uint32_t)buff[34] << 16) +
		((uint32_t)buff[35] << 24);
	if (lba0 == 0) {
		fprintf(stderr, "Big sectors is 0: unformatted SD?\n");
		return -1;
	}
	return 0;
}

int boot_to_mbr(int fd_out)
{
	int bytes_written;

	lseek(fd_out, 0, SEEK_SET);
	memcpy(mbr + 90, boot, sizeof(boot));
	bytes_written = write(fd_out, mbr, sizeof(mbr));
	if (bytes_written != (int)sizeof(mbr))
		return -1;
	return 0;
}

int copy_to_image(int fd_out, const char *filename, int lba)
{
	int fd_in;
	int total;

	printf("Copying %s to SD image at LBA %d\n", filename, lba);
	fd_in = open(filename, O_RDONLY);
	if (fd_in < 0) {
		perror(filename);
		return -1;
	}

	lseek(fd_out, (off_t)512 * lba, SEEK_SET);
	total = 0;
	for (;;) {
		int bytes_read, bytes_written;
		bytes_read = read(fd_in, buff, sizeof(buff));
		if (bytes_read < 0) {
			if (errno == EAGAIN)
				continue;
			perror("read()");
			return -1;
		}
		if (bytes_read == 0)
			break;
		bytes_written = write(fd_out, buff, bytes_read);
		if (bytes_written < 0) {
			if (errno == EAGAIN) {
				lseek(fd_in, -bytes_read, SEEK_CUR);
				continue;
			}
			perror("write()");
			return -1;
		}
		total += bytes_written;
	}
	return total;
}

void usage(int argc, char **argv)
{
	char *program;

	program = strrchr(argv[0], '/');
	if (program)
		program++;
	else
		program = argv[0];
	printf("Usage: %s [options]\n", program);
	printf("Options can be one or more of:\n");
	printf("-s	select device to write to (e.g. /dev/sdb1)\n");
	printf("-b	select boot image (default: %s)\n", bootrom);
	printf("-n	set noname (default: %s)\n", nodename);
	printf("-da	set filename of the A: image (default: %s)\n", disk_a);
	printf("-db	set filename of the A: image (default: %s)\n", disk_a);
	printf("-dc	set filename of the A: image (default: %s)\n", disk_a);
	printf("-dd	set filename of the A: image (default: %s)\n", disk_a);
	printf("-de	set filename of the A: image (default: %s)\n", disk_a);
	printf("-df	set filename of the A: image (default: %s)\n", disk_a);
	printf("-dg	set filename of the A: image (default: %s)\n", disk_a);
	printf("-dh	set filename of the A: image (default: %s)\n", disk_a);
}

int main(int argc, char **argv)
{
	int i;
	int fd;
	int lba;

	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-s")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -s\n");
				return 1;
			}
			snprintf(devicename, sizeof(devicename), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-n")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing node name for -n\n");
				return 1;
			}
			snprintf(nodename, sizeof(nodename), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-b")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -b\n");
				return 1;
			}
			snprintf(disk_a, sizeof(disk_a), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-da")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -da\n");
				return 1;
			}
			snprintf(disk_a, sizeof(disk_a), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-db")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -db\n");
				return 1;
			}
			snprintf(disk_b, sizeof(disk_b), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-dc")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -dc\n");
				return 1;
			}
			snprintf(disk_c, sizeof(disk_c), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-dd")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -dd\n");
				return 1;
			}
			snprintf(disk_d, sizeof(disk_d), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-de")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -de\n");
				return 1;
			}
			snprintf(disk_e, sizeof(disk_e), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-df")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -df\n");
				return 1;
			}
			snprintf(disk_f, sizeof(disk_f), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-dg")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -dg\n");
				return 1;
			}
			snprintf(disk_g, sizeof(disk_g), "%s", argv[i]);
			continue;
		}
		if (!strcmp(argv[i], "-dh")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing pathname for -dh\n");
				return 1;
			}
			snprintf(disk_h, sizeof(disk_h), "%s", argv[i]);
			continue;
		}
		fprintf(stderr, "Unknow option: %s\n", argv[i]);
		usage(argc, argv);
		return 1;
	}

	if (strlen(devicename) == 0) {
		fprintf(stderr, "No device name selected\n");
		usage(argc, argv);
		return 1;
	}

	fd = open(devicename, O_RDWR | O_BINARY);
	i = get_base_lba(fd);
	if (i < 0)
		return 1;
	memcpy(mbr, buff, sizeof(mbr));

	printf("First sector after DOS/FAT region is: %d\n", lba0);
	i = copy_to_image(fd, bootrom, lba0);
	if (i < 0)
		return 1;
	printf("\tcopied %d bytes\n", i);
	lba = lba0;						/* overwrite with the first image data */
	memcpy(boot, buff, sizeof(boot));			/* copy the data to the boot image */
	snprintf((char *)boot + 0xd0, 16, "%s", nodename);	/* write the node name at offset 0xd0 */

	i = copy_to_image(fd, disk_a, lba);
	if (i < 0)
		return 1;
	printf("\tcopied %d bytes\n", i);
	write_long(boot, 0xe0 + 0*4, lba);
	lba = lba + (i + 511) / 512;

	i = copy_to_image(fd, disk_b, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 1*4, lba);
		lba = lba + (i + 511) / 512;
	}

	i = copy_to_image(fd, disk_c, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 2*4, lba);
		lba = lba + (i + 511) / 512;
	}

	i = copy_to_image(fd, disk_d, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 3*4, lba);
		lba = lba + (i + 511) / 512;
	}

	i = copy_to_image(fd, disk_e, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 4*4, lba);
		lba = lba + (i + 511) / 512;
	}

	i = copy_to_image(fd, disk_f, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 5*4, lba);
		lba = lba + (i + 511) / 512;
	}

	i = copy_to_image(fd, disk_g, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 6*4, lba);
		lba = lba + (i + 511) / 512;
	}

	i = copy_to_image(fd, disk_h, lba);
	if (i > 0) {
		printf("\tcopied %d bytes\n", i);
		write_long(boot, 0xe0 + 7*4, lba);
		lba = lba + (i + 511) / 512;
	}

	printf("Updating DOS/FAT boot sector at %d\n", 0);
	i = boot_to_mbr(fd);
	if (i < 0) {
		fprintf(stderr, "Failed! The SD may be unusable for CP/M\n");
		return 1;
	}
	close(fd);

	return 0;
}
